/*
 * @(#)FormHConstraints.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.constraints;

import java.awt.Container;

import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.List;

import java.util.StringTokenizer;

import com.jgoodies.forms.layout.CellConstraints;
import com.jgoodies.forms.layout.FormLayout;

import chrriis.uihierarchy.SingleLevelAccessor;

/**
 * The constraints for a Form layout (www.jgoodies.com). Constraints are
 * manipulated by the <code>createConstraints</code> method, if they are a
 * String.
 * <br><br>
 * The format is directly what is expected by the form layout.<br>
 * <br>
 * The only difference is that the constraints for the parent (the rows and
 * the columns groups) can be defined using a String representation of the
 * form: <code>"rowgroups=[[1, 2], [3, 4]], columngroups=[[1, 2],
 * [3, 4]]"</code>.<br>
 * Several arrays can be used (not only two as in the example), and several
 * comma-separated numbers. Also, it is possible to constrain only
 * <code>rowgroups</code> or <code>columngroups</code>.
 * @version 1.0 2003.10.23
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class FormHConstraints implements LayoutHConstraints {

  /**
   * Create the constraints for a Form layout, if the parameter is a String.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param parentContainer The parent container.
   * @param constraints The constraints of the component.
   * @return The cell constraints, the original constraints if not a String, or
   * null if the parent container is the component.
   */
  public Object createConstraints(SingleLevelAccessor accessor, Container parentContainer, Object constraints) {
    if(!(constraints instanceof String)) {
      return constraints;
    }
    if(accessor.getComponent() == parentContainer) {
      String cons = ((String)constraints).toLowerCase();
      int rowGroups = cons.indexOf("rowgroups");
      int columnGroups = cons.indexOf("columngroups");
      String groupsConstraints = null;
      if(rowGroups != -1) {
        if(columnGroups > rowGroups) {
          groupsConstraints = cons.substring(rowGroups, columnGroups);
        } else {
          groupsConstraints = cons.substring(rowGroups);
        }
        int[][] groups = extractGroups(groupsConstraints);
        ((FormLayout)parentContainer.getLayout()).setRowGroups(groups);
      }
      if(columnGroups != -1) {
        if(rowGroups > columnGroups) {
          groupsConstraints = cons.substring(columnGroups, rowGroups);
        } else {
          groupsConstraints = cons.substring(columnGroups);
        }
        int[][] groups = extractGroups(groupsConstraints);
        ((FormLayout)parentContainer.getLayout()).setColumnGroups(groups);
      }
      return null;
    }
    return new CellConstraints((String)constraints);
  }

  /**
   * Get the constraints for a groups description.
   * @param groupsConstraints The groups description to parse.
   * @return The extracted constraints.
   */
  protected int[][] extractGroups(String groupsConstraints) {
    groupsConstraints = groupsConstraints.substring(groupsConstraints.indexOf("[") + 1, groupsConstraints.lastIndexOf("]"));
    List groupList = new ArrayList(1); 
    while(true) {
      int end = groupsConstraints.indexOf("]");
      if(end == -1) {
        break;
      }
      String group = groupsConstraints.substring(groupsConstraints.indexOf("[") + 1, end);
      StringTokenizer st = new StringTokenizer(group, ",");
      int[] tokens = new int[st.countTokens()];
      for(int i=0; i<tokens.length; i++) {
        tokens[i] = Integer.parseInt(st.nextToken().trim());
      }
      groupList.add(tokens);
      groupsConstraints = groupsConstraints.substring(end + 1);
    }
    int[][] groups = new int[groupList.size()][];
    for(int i=0; i<groupList.size(); i++) {
      groups[i] = (int[])groupList.get(i);
    }
    return groups;
  }

}
