/*
 * @(#)SwingUtilities_1_!.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.compatibility;

import java.awt.AWTEvent;
import java.lang.reflect.InvocationTargetException;

/**
 * The Java 1.1 implementation of SwingUtilities.
 * @version 1.0 2004.01.11
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
class SwingUtilities_1_1 {

  /** The event dispatch thread. */
  protected static Class eventDispatchThreadClass;

  /**
   * Indicate if the current thread is the event dispatch thread.
   * @return True if the current thread is the dispatch thread.
   */
  public static boolean isEventDispatchThread() {
    if(!Compatibility.isSwingPresent()) {
      Thread currentThread = Thread.currentThread();
      if(eventDispatchThreadClass == null) {
        Class currentThreadClass = currentThread.getClass();
        if(currentThreadClass.getName().indexOf("EventDispatchThread") >= 0 || currentThreadClass.getName().indexOf("JMEventQueue") >= 0) {
          eventDispatchThreadClass = currentThreadClass;
          return true;
        }
        return false;
      }
      return eventDispatchThreadClass.isInstance(currentThread);
    }
    return javax.swing.SwingUtilities.isEventDispatchThread();
  }

  /**
   * Runs a Runnable in the dispatch thread, synchronously.
   * @param runnable The action to run.
   * @throws InterruptedException If interrupted.
   * @throws InvocationTargetException If exception on target.
   */
  public static void invokeAndWait(final Runnable runnable) throws InterruptedException, InvocationTargetException {
    if(!Compatibility.isSwingPresent()) {
      if(isEventDispatchThread()) {
        throw new Error("Cannot call invokeAndWait from the event dispatcher thread");
      }
      Exception exception = postRunnable(runnable, true);
      if(exception != null) {
        throw new InvocationTargetException(exception);
      }
    } else {
      javax.swing.SwingUtilities.invokeAndWait(runnable);
    }
  }

  /**
   * Run a Runnable in the event dispatch thread, asynchronously.
   * @param runnable The action to run.
   */
  public static void invokeLater(Runnable runnable) {
    try {
      postRunnable(runnable, false);
    } catch(Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * A runnable event.
   */
  protected static class RunnableEvent extends AWTEvent {
    static final int EVENT_ID = 2999;
    static final java.awt.Component TARGET = new java.awt.Component() {
      {
        enableEvents(EVENT_ID);
      }
      protected void processEvent(AWTEvent awtevent) {
        if(awtevent instanceof RunnableEvent) {
          processRunnableEvent((RunnableEvent)awtevent);
        }
      }
    };
    final Runnable runnable;
    final Object lock;
    boolean printExceptions;
    Exception exception;
    RunnableEvent(Runnable runnable, Object obj, boolean printExceptions) {
      super(TARGET, EVENT_ID);
      this.runnable = runnable;
      lock = obj;
      this.printExceptions = printExceptions;
    }
  }

  /**
   * Process a runnable event.
   * @param runnableEvent
   */
  protected static void processRunnableEvent(RunnableEvent runnableEvent) {
    Object obj = runnableEvent.lock;
    if(obj == null) {
      runnableEvent.runnable.run();
    } else {
      synchronized(obj) {
        try {
          runnableEvent.runnable.run();
        } catch(Exception ex) {
          runnableEvent.exception = ex;
          if(runnableEvent.printExceptions) {
            runnableEvent.exception.printStackTrace();
          }
        } finally {
          if(runnableEvent.lock != null) {
            runnableEvent.lock.notify();
          }
        }
      }
    }
  }

  /**
   * Post a runnable.
   * @param runnable The runnable to run.
   * @param isWaiting True if the posting should wait.
   * @return The exception that was raised, or null.
   * @throws InterruptedException If interrupted.
   */
  protected static Exception postRunnable(Runnable runnable, boolean isWaiting) throws InterruptedException {
    Object obj = new Object();
    synchronized(obj) {
      java.awt.EventQueue eventqueue = null;
      try {
        eventqueue = java.awt.Toolkit.getDefaultToolkit().getSystemEventQueue();
      } catch(SecurityException _ex) {}
      RunnableEvent runnableEvent = new RunnableEvent(runnable, obj, !isWaiting);
      if(eventqueue != null) {
        eventqueue.postEvent(runnableEvent);
        if(isWaiting) {
          obj.wait();
        }
      } else {
        processRunnableEvent(runnableEvent);
      }
      return runnableEvent.exception;
    }
  }

}
