/*
 * @(#)SpringHConstraints.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.constraints;

import java.awt.Container;

import javax.swing.Spring;
import javax.swing.SpringLayout;

import chrriis.uihierarchy.SingleLevelAccessor;
import chrriis.uihierarchy.util.HStringTokenizer;

/**
 * The constraints for a Spring layout. Constraints are manipulated by the
 * <code>createConstraints</code> method, if they are a String.
 * <br><br>
 * The format is of the form "key1=expression1, key2=expression2, ...".
 * <br>
 * The possible keys are a direction or a location: "west", "north", "east",
 * "south", "x", "y", "width" and "height".
 * <br>
 * The expressions are combinations of constants and component based
 * constraints. An example can be: "5+comp.east", where "comp" is a component
 * that was mapped to this name. Additionaly, there is support for
 * max(a, b, ...) with parameters that are themselves expressions.<br>
 * With the Spring layout, the container itself needs to get some constraints.
 * It is possible to create the constraints for the container using the
 * <code>subConstrain</code> method.
 * @version 1.1 2003.11.03
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class SpringHConstraints implements LayoutHConstraints {

  /**
   * Create the constraints for a Spring layout, if the parameter is a String.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param parentContainer The parent container.
   * @param constraints The constraints of the component.
   * @return null, or the original constraints if not a String.
   * @see chrriis.uihierarchy.HNode#subConstrain(java.lang.Object)
   * @see chrriis.uihierarchy.HRootNode#subConstrain(java.lang.Object)
   * @see chrriis.uihierarchy.HLeaf#map(java.lang.String)
   * @see chrriis.uihierarchy.HNode#map(java.lang.String)
   * @see chrriis.uihierarchy.HRootNode#map(java.lang.String)
   */
  public Object createConstraints(SingleLevelAccessor accessor, Container parentContainer, Object constraints) {
    if(!(constraints instanceof String)) {
      return constraints;
    }
    HStringTokenizer[] tokenizers = new HStringTokenizer((String)constraints).getSubTokenizers();
    SpringLayout.Constraints layoutConstraints = ((SpringLayout)parentContainer.getLayout()).getConstraints(accessor.getComponent());
    SpringLayout layout = (SpringLayout)parentContainer.getLayout();
    for(int i=0; i<tokenizers.length; i++) {
      HStringTokenizer st = tokenizers[i];
      char operator = st.getOperator();
      if(operator != '=') {
        throw new IllegalStateException("Constraints must be key/expression pairs!");
      }
      HStringTokenizer[] subTokenizers = st.getSubTokenizers();
      String left = subTokenizers[0].getToken().toLowerCase();
      if(left.equals("west")) {
        layoutConstraints.setConstraint(SpringLayout.WEST, createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("north")) {
        layoutConstraints.setConstraint(SpringLayout.NORTH, createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("east")) {
        layoutConstraints.setConstraint(SpringLayout.EAST, createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("south")) {
        layoutConstraints.setConstraint(SpringLayout.SOUTH, createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("x")) {
        layoutConstraints.setX(createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("y")) {
        layoutConstraints.setY(createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("width")) {
        layoutConstraints.setWidth(createSpring(accessor, layout, subTokenizers[1]));
      } else if(left.equals("height")) {
        layoutConstraints.setHeight(createSpring(accessor, layout, subTokenizers[1]));
      } else {
        throw new IllegalStateException("\"" + subTokenizers[0].getToken() + "\" is not a valid key!");
      }
    }
    return null;
  }

  /**
   * Create a Spring from a hierarchy tokenizer representation.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param layout The spring layout to use for the constraints.
   * @param tokenizer The hierarchy tokenizer representing the constraints.
   * @return The Spring created from the tokenizer.
   */
  protected Spring createSpring(SingleLevelAccessor accessor, SpringLayout layout, HStringTokenizer tokenizer) {
    char operator = tokenizer.getOperator();
    String token = tokenizer.getToken();
    HStringTokenizer[] sts = tokenizer.getSubTokenizers();
    switch(operator) {
      case '-':
        if(sts[0] == null) {
          return Spring.minus(createSpring(accessor, layout, sts[1]));
        }
        return Spring.sum(createSpring(accessor, layout, sts[0]), Spring.minus(createSpring(accessor, layout, sts[1])));
      case '+':
        return Spring.sum(createSpring(accessor, layout, sts[0]), createSpring(accessor, layout, sts[1]));
      case '(':
        if(token != null) {
          if(!token.toLowerCase().equals("max")) {
            throw new IllegalStateException("\"" + token + "\" is not a valid function!");
          }
          if(sts.length < 1) {
            throw new IllegalStateException("\"" + token + "\" function must have one or more parameters!");
          }
          if(sts.length == 1) {
            return createSpring(accessor, layout, sts[0]);
          }
          Spring sum = Spring.max(createSpring(accessor, layout, sts[0]), createSpring(accessor, layout, sts[1]));
          for(int i=0; i<sts.length-1; i++) {
            sum = Spring.max(sum, createSpring(accessor, layout, sts[i + 1]));
          }
          return sum;
        }
        if(sts.length != 1) {
          throw new IllegalStateException("Braces must have one and only one expression!");
        }
        return createSpring(accessor, layout, sts[0]);
      case 0:
        if(token == null) {
          throw new IllegalStateException("A group must be composed of one and only one expression!");
        }
        int index = token.indexOf(".");
        if(index > 0) {
          String left = token.substring(0, index);
          String right = token.substring(index+1).toLowerCase();
          if(right.equals("west")) {
            right = SpringLayout.WEST;
          } else if(right.equals("north")) {
            right = SpringLayout.NORTH;
          } else if(right.equals("east")) {
            right = SpringLayout.EAST;
          } else if(right.equals("south")) {
            right = SpringLayout.SOUTH;
          }
          return layout.getConstraints(((SingleLevelAccessor)accessor.getConfig().createAccessor(left)).getComponent()).getConstraint(right);
        }
        return Spring.constant(Integer.parseInt(token));
      default:
        throw new IllegalStateException("Cannot handle operator \"" + operator + "\"!");
    }
  }

}
