/*
 * @(#)HLevel.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy;

import java.awt.Component;
import java.awt.Container;

import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.List;
import chrriis.uihierarchy.creator.ComponentCreator;

/**
 * A level in the hierarchy.
 * @version 2.0 2003.11.02
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public abstract class HLevel {

  /** The configuration. */
  protected UIHConfig uihConfig;

  /** The component. */
  protected Component component;

  /** The parent. */
  protected HParentLevel parent;

  /**
   * Construct a hierarchy level.
   * @param uihConfig The configuration.
   * @param parent The hierarchy parent level of this one.
   * @param component The component.
   */
  protected HLevel(UIHConfig uihConfig, HParentLevel parent, Object component) {
    this.uihConfig = uihConfig;
    this.parent = parent;
    setComponent(component);
    if(parent != null) {
      parent.addChild(this);
    }
  }

  /**
   * Set the component by creating one from an object component.
   * @param component The component object.
   */
  protected void setComponent(Object component) {
    if(component instanceof Component) {
      this.component = (Component)component;
    } else {
      ComponentCreator creator;
      UIHConfig config = getConfig();
      if(component instanceof ComponentCreator) {
        creator = (ComponentCreator)component;
      } else {
        creator = config.getComponentCreator(component == null? null: component.getClass());
      }
      int level;
      if(this instanceof HLeaf) {
        level = ComponentCreator.LEAF;
      } else {
        level = ComponentCreator.NODE;
      }
      if(creator == null) {
        throw new UIHException(config.createAccessor(this), "The component must be of class Component or have a matching ComponentCreator!");
      }
      Container parentContainer = null;
      if(!(this instanceof HRootNode)) {
        HParentLevel parentLevel = getParent();
        if(parentLevel != null) {
          parentContainer = parentLevel.getContainer();
        }
      }
      this.component = creator.createComponent(config, parentContainer, component, level);
    }
    if(!(this instanceof HRootNode) && this.component.getParent() != null) {
      throw new UIHException(getConfig().createAccessor(this), "The component must not already have a parent! If you move some levels, use an accessor and add a new hierarchy to it.");
    }
  }

  /**
   * Get the configuration.
   * @return The configuration.
   */
  protected UIHConfig getConfig() {
    return uihConfig;
  }

  /**
   * Get the component at this hierarchy level.
   * @return The component.
   */
  protected Component getComponent() {
    return component;
  }

  /**
   * Get the parent.
   * @return The parent.
   */
  protected HParentLevel getParent() {
    return parent;
  }

/*----------------------------------------------------------------------------*/

  /** The constraints, if any. */
  protected Object constraints;

  /**
   * Get the constraints.
   * @return The constraints.
   */
  protected Object getConstraints() {
    return constraints;
  }

  /**
   * Set the constraints.
   * @param constraints The constraints.
   */
  protected void constrain_(Object constraints) {
    this.constraints = constraints;
    clearRealizedConstraints();
  }

  /** The constraints used for heritance, if set. */
  protected Object realizedConstraints;

  /**
   * Get the realized constraints.
   * @return The realized constraints.
   */
  protected Object getRealizedConstraints() {
    return realizedConstraints;
  }

  /**
   * Set the realized constraints.
   * @param realizedConstraints The realized constraints.
   */
  protected void setRealizedConstraints(Object realizedConstraints) {
    this.realizedConstraints = realizedConstraints;
  }

  /**
   * Clear the constraints and the realized constraints.
   */
  protected void clearConstraints() {
    constraints = null;
    clearRealizedConstraints();
  }

  /**
   * Clear the realized constraints.
   */
  protected void clearRealizedConstraints() {
    realizedConstraints = null;
  }

/*----------------------------------------------------------------------------*/

  /** The number of slots. */
  protected static int slotCount = 0;

  /**
   * Aquire the next free slot.
   * @return The next free slot.
   */
  protected static int aquireSlot() {
    return slotCount++;
  }

  /**
   * Get the number of used slots.
   * @return The number of slots.
   */
  protected int getSlotCount() {
    return slotCount;
  }

  /** The list of objects in the slots. */
  protected List slotObjectsList;

  /**
   * Check that the given slot is valid.
   * @param slot The slot to check.
   */
  protected void checkSlot(int slot) {
    int slotCount = getSlotCount();
    if(slot < 0) {
      throw new IndexOutOfBoundsException("The slot index cannot be negative!");
    } else if (slot >= slotCount){
      throw new IndexOutOfBoundsException("The slot index cannot be greater or equal than the slot count (" + slotCount + ")!");
    }
  }

  /**
   * Get the object at the particular slot.
   * @param slot The slot from where to get the object.
   * @return The object that was stored at the indicated slot.
   */
  protected Object getSlotObject(int slot) {
    checkSlot(slot);
    if(slotObjectsList == null || slotObjectsList.size() <= slot) {
      return null;
    }
    return slotObjectsList.get(slot);
  }

  /**
   * Set the object at a given slot.
   * @param slotObject The object to set.
   * @param slot The slot at which to set the object.
   */
  protected void setSlotObject(Object slotObject, int slot) {
    checkSlot(slot);
    if(slotObject == null) {
      if(slotObjectsList != null && slotObjectsList.size() >= slot) {
        slotObjectsList.set(slot, slotObject);
      }
    } else {
      if(slotObjectsList == null) {
        slotObjectsList = new ArrayList(slot + 1);
      } else {
        ((ArrayList)slotObjectsList).ensureCapacity(slot + 1);
      }
      while(slotObjectsList.size() <= slot) {
        slotObjectsList.add(null);
      }
      slotObjectsList.set(slot, slotObject);
    }
  }

/*----------------------------------------------------------------------------*/

  /**
   * Map the given name to the current level.
   * @param name The name of the level.
   */
  protected void map_(String name) {
    getConfig().map(this, name);
  }

/*----------------------------------------------------------------------------*/

  /**
   * Add the level to the debugging process.
   * @param parameters The parameters to use with the debugger.
   */
  protected void debug_(Object parameters) {
    getConfig().addLevelDebugging(this, parameters);
  }

/*----------------------------------------------------------------------------*/

  /**
   * Get the name of a class, without the package name. This is for example
   * useful when printing the hierarchy to reduce clutter.
   * @param classObject The class for which to get the name.
   * @return The name of the class without the package name.
   */
  protected String getClassName(Class classObject) {
    String name = classObject.getName();
    return name.substring(name.lastIndexOf(".") + 1);
  }

/*----------------------------------------------------------------------------*/

  /** Indicate if the level is realized. */
  protected boolean isRealized = false;

  /**
   * Indicate if the level is realized.
   * @return True if the level is realized.
   */
  protected boolean isRealized() {
    return isRealized;
  }

  /**
   * Set whether the level is realized.
   * @param isRealized True if the level is to be considered realized.
   */
  protected void setRealized(boolean isRealized) {
    this.isRealized = isRealized;
  }

  /**
   * Realize the level.
   */
  protected void realize() {
    isRealized = true;
  }

/*----------------------------------------------------------------------------*/

  /**
   * Remove the level from its parent.
   */
  protected void remove() {
    HParentLevel hParentLevel = getParent();
    if(hParentLevel != null) {
      hParentLevel.removeChild(this);
    }
  }

/*----------------------------------------------------------------------------*/

  /**
   * Get the level path as a String dotted notation.
   * @return The String dotted notation of the path to this level.
   */
  protected String getDottedDecimalPath() {
    HParentLevel hParentLevel = getParent();
    if(hParentLevel == null) {
      return null;
    }
    List childrenList = hParentLevel.getChildrenList();
    int index = childrenList.indexOf(this) + 1;
    if(index == 0) {
      index = childrenList.size() + 1;
    }
    String parentDottedDecimalPath = hParentLevel.getDottedDecimalPath();
    if(parentDottedDecimalPath != null) {
      return parentDottedDecimalPath + "." + index;
    }
    return String.valueOf(index);
  }

}
