/*
 * @(#)GridBagHConstraints.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.constraints;

import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.StringTokenizer;

import chrriis.uihierarchy.SingleLevelAccessor;
import chrriis.uihierarchy.UIHConfig;
import chrriis.uihierarchy.util.UnitParser;

/**
 * The constraints for a GridBag layout. Constraints are manipulated by the
 * <code>createConstraints</code> method, if they are a String.
 * <br><br>
 * The format is of the form "inherit, key1=value1, key2=value2, ...". Inherit
 * is optional and can have two forms: "inherit" or "inherit=mappedName". The
 * inherit keyword is used to first apply some earlier constraints before
 * applying a new set of constraints. This allows chaining of constraints.
 * If used with a mapped name then it will use the constraints applied to the
 * level that was mapped, else it will use the previous sibbling's
 * constraints.<br>
 * Additionaly, "gridx" and "gridy" accept the special forms "gridx++" and
 * "gridy++" which take the value of the previous level, incremented by one.<br>
 * <br>
 * Key/value pairs are (case is insensitive):
 * <br>
 * <ul>
 * <li>gridx = an integer.</li>
 * <li>gridy = an integer.</li>
 * <li>gridwidth = an integer, or one of the possible GridBagConstraints fields
 * (e.g.: REMAINDER).</li>
 * <li>gridheight = an integer, or one of the possible GridBagConstraints fields
 * (e.g.: REMAINDER).</li>
 * <li>weightx = a double.</li>
 * <li>weighty = a double.</li>
 * <li>anchor = one of the possible GridBagConstraints fields (e.g.: NORTH).
 * </li>
 * <li>fill = one of the possible GridBagConstraints fields (e.g.: BOTH).</li>
 * <li>insets = four dimensions in the form "top:left:bottom:right".</li>
 * <li>ipadx = an integer.</li>
 * <li>ipady = an integer.</li>
 * </ul>
 * @version 1.1 2003.11.03
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class GridBagHConstraints implements LayoutHConstraints {

  /**
   * Create the constraints for a gridbag layout, if the parameter is a String.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param parentContainer The parent container.
   * @param constraints The constraints of the component.
   * @return null, or the original constraints if not a String.
   */
  public Object createConstraints(SingleLevelAccessor accessor, Container parentContainer, Object constraints) {
    if(!(constraints instanceof String)) {
      return (GridBagConstraints)constraints;
    }
    GridBagConstraints gridBagConstraints = analyzeConstraints(accessor, constraints);
    ((GridBagLayout)parentContainer.getLayout()).setConstraints(accessor.getComponent(), gridBagConstraints);
    return null;
  }

  /**
   * Analize the constraints contained in a String representation.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param constraints The constraints in a String form.
   * @param accessor The accessor of this level.
   * @return The grid bag constraints corresponding to the String
   * representation.
   */
  protected GridBagConstraints analyzeConstraints(SingleLevelAccessor accessor, Object constraints) {
    if(!(constraints instanceof String)) {
      return (GridBagConstraints)constraints;
    }
    GridBagConstraints gridBagConstraints = (GridBagConstraints)accessor.getConstraintsHandler().getRealizedConstraints();
    if(gridBagConstraints != null) {
      return gridBagConstraints;
    }
    gridBagConstraints = new GridBagConstraints();
    StringTokenizer st = new StringTokenizer((String)constraints, ",");
    boolean isFirstToken = true;
    SingleLevelAccessor previousAccessor = null;
    UIHConfig uihConfig = accessor.getConfig();
    while(st.hasMoreTokens()) {
      String token = st.nextToken().trim();
      int index = token.indexOf("=");
      if(index != -1) {
        String oKey = token.substring(0, index).trim();
        String key = oKey.toLowerCase();
        String oValue = token.substring(index + 1).trim();
        String value = oValue.toLowerCase();
        boolean isKeyOK = true;
        if(key.equals("inherit")) {
          if(!isFirstToken) {
            throw new IllegalArgumentException("The \"inherit\" constraint can only be used as a first constraint!");
          }
          SingleLevelAccessor valueAccessor = (SingleLevelAccessor)uihConfig.createAccessor(oValue);
          GridBagConstraints gridBagHeritedConstraints = analyzeConstraints(accessor, valueAccessor.getConstraintsHandler().getConstraints());
          if(gridBagHeritedConstraints == null) {
            throw new IllegalArgumentException("\"" + oValue + "\" is not a possible value for \"" + oKey + "\"!");
          }
          gridBagConstraints = (GridBagConstraints)gridBagHeritedConstraints.clone();
        } else {
          try {
            if(key.equals("gridx")) {
              gridBagConstraints.gridx = Integer.parseInt(value);
            } else if(key.equals("gridy")) {
              gridBagConstraints.gridy = Integer.parseInt(value);
            } else if(key.equals("gridwidth")) {
              try {
                gridBagConstraints.gridwidth = Integer.parseInt(value);
              } catch(Exception e) {
                gridBagConstraints.gridwidth = ((Integer)GridBagConstraints.class.getField(value.toUpperCase()).get(null)).intValue();
              }
            } else if(key.equals("gridheight")) {
              try {
                gridBagConstraints.gridheight = Integer.parseInt(value);
              } catch(Exception e) {
                gridBagConstraints.gridheight = ((Integer)GridBagConstraints.class.getField(value.toUpperCase()).get(null)).intValue();
              }
            } else if(key.equals("weightx")) {
              gridBagConstraints.weightx = Double.valueOf(value).doubleValue();
            } else if(key.equals("weighty")) {
              gridBagConstraints.weighty = Double.valueOf(value).doubleValue();
            } else if(key.equals("anchor")) {
              gridBagConstraints.anchor = ((Integer)GridBagConstraints.class.getField(value.toUpperCase()).get(null)).intValue();
            } else if(key.equals("fill")) {
              gridBagConstraints.fill = ((Integer)GridBagConstraints.class.getField(value.toUpperCase()).get(null)).intValue();
            } else if(key.equals("insets")) {
              StringTokenizer st2 = new StringTokenizer(value, ":");
              gridBagConstraints.insets = new Insets(UnitParser.intPixelValue(uihConfig, st2.nextToken()),
                                                     UnitParser.intPixelValue(uihConfig, st2.nextToken()),
                                                     UnitParser.intPixelValue(uihConfig, st2.nextToken()),
                                                     UnitParser.intPixelValue(uihConfig, st2.nextToken()));
            } else if(key.equals("ipadx")) {
              gridBagConstraints.ipadx = Integer.parseInt(value);
            } else if(key.equals("ipady")) {
              gridBagConstraints.ipady = Integer.parseInt(value);
            } else {
              isKeyOK = false;
            }
          } catch(Exception e) {
            throw new IllegalArgumentException("\"" + oValue + "\" is not a possible value for \"" + oKey + "\"!");
          }
        }
        if(!isKeyOK) {
          throw new IllegalArgumentException("\"" + oKey + "\" is not a known constraint key!");
        }
      } else {
        String key = token.toLowerCase();
        if(key.equals("inherit")) {
          if(!isFirstToken) {
            throw new IllegalArgumentException("The \"inherit\" constraint can only be used as a first constraint!");
          }
          try {
            if(previousAccessor == null) {
              previousAccessor = accessor.previous();
            }
            GridBagConstraints gridBagHeritedConstraints = analyzeConstraints(previousAccessor, previousAccessor.getConstraintsHandler().getConstraints());
            gridBagConstraints = (GridBagConstraints)gridBagHeritedConstraints.clone();
          } catch(Exception e) {
            String message = e.getMessage();
            message = message != null? " " + message: "";
            throw new IllegalArgumentException("Failed to \"inherit\"." + message);
          }
        } else if(key.startsWith("gridx") || key.startsWith("gridy")) {
          if(previousAccessor == null) {
            previousAccessor = accessor.previous();
          }
          key = key.substring("gridx".length()).trim();
          if(key.equals("++")) {
            if(previousAccessor == null) {
              throw new IllegalStateException("There is no previous level from which to increment " + token.substring(0, "gridx".length()) + "!");
            }
            GridBagConstraints gridBagHeritedConstraints = null;
            try {
              gridBagHeritedConstraints = analyzeConstraints(previousAccessor, previousAccessor.getConstraintsHandler().getConstraints());
            } catch(Exception e) {}
            if(token.toLowerCase().charAt("gridx".length() - 1) == 'x') {
              gridBagConstraints.gridx = gridBagHeritedConstraints == null? 1: Math.max(1, gridBagHeritedConstraints.gridx + 1);
            } else {
              gridBagConstraints.gridy = gridBagHeritedConstraints == null? 1: Math.max(1, gridBagHeritedConstraints.gridy + 1);
            }
          } else {
            throw new IllegalArgumentException("\"" + token + "\" is not a known constraint!");
          }
        } else {
          throw new IllegalArgumentException("\"" + token + "\" is not a known constraint!");
        }
      }
      isFirstToken = false;
    }
    accessor.getConstraintsHandler().setRealizedConstraints(gridBagConstraints);
    return gridBagConstraints;
  }

}
