/*
 * @(#)NullHConstraints.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.constraints;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Point;

import chrriis.uihierarchy.compatibility.HashMap;
import chrriis.uihierarchy.compatibility.Map;
import chrriis.uihierarchy.util.UnitParser;

import java.util.StringTokenizer;

import chrriis.uihierarchy.SingleLevelAccessor;
import chrriis.uihierarchy.UIHConfig;

/**
 * The constraints for a null layout. Constraints are manipulated by the
 * <code>createConstraints</code> method, if they are a String.
 * Additionaly, the size of the component is computed if it is not yet set.
 * <br><br>
 * The format is of the form "key1=value1, key2=value2, ..." where pairs are:
 * <br>
 * <ul>
 * <li>x = an integer that specifies the absolute x location.
 * <li>y = an integer that specifies the absolute y location.
 * <li>width = an integer that specifies the width.
 * <li>height = an integer that specifies the height.
 * </ul>
 * <br>
 * If the component is not displayable and its size is (0, 0), then it will be
 * set to its preffered size.
 * @version 1.0 2003.11.04
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class NullHConstraints implements LayoutHConstraints {

  /**
   * Create the constraints for a null layout, if the parameter is a String.
   * @param accessor The accessor of the level for which to create the
   * constraints.
   * @param parentContainer The parent container.
   * @param constraints The constraints of the component.
   * @return null, or the original constraints if not a String.
   */
  public Object createConstraints(SingleLevelAccessor accessor, Container parentContainer, Object constraints) {
    Component component = accessor.getComponent();
    Dimension size = component.getSize();
    Point location = component.getLocation();
    if(!chrriis.uihierarchy.compatibility.Component.isDisplayable(component) && size.equals(new Dimension(0, 0))) {
      size = component.getPreferredSize();
      component.setSize(size);
    }
    if(!(constraints instanceof String)) {
      return constraints;
    }
    Map constraintsMap = new HashMap(4);
    StringTokenizer st = new StringTokenizer((String)constraints, ",");
    UIHConfig uihConfig = accessor.getConfig();
    while(st.hasMoreTokens()) {
      String token = st.nextToken();
      int index = token.indexOf("=");
      if(index != -1) {
        String key = token.substring(0, index).trim().toLowerCase();
        String value = token.substring(index + 1).trim();
        if(key.equals("x") || key.equals("y") || key.equals("width") || key.equals("height")) {
//          constraintsMap.put(key, new Integer(value));
          constraintsMap.put(key, new Integer(UnitParser.intPixelValue(uihConfig, value)));
        }
      }
    }
    try {
      int x;
      Integer value = (Integer)constraintsMap.get("x");
      if(value != null) {
        x = value.intValue();
      } else {
        x = location.x;
      }
      int y;
      value = (Integer)constraintsMap.get("y");
      if(value != null) {
        y = value.intValue();
      } else {
        y = location.y;
      }
      component.setLocation(x, y);
      int width;
      value = (Integer)constraintsMap.get("width");
      if(value != null) {
        width = value.intValue();
      } else {
        width = size.width;
      }
      int height;
      value = (Integer)constraintsMap.get("height");
      if(value != null) {
        height = value.intValue();
      } else {
        height = size.height;
      }
      component.setSize(width, height);
    } catch(Exception e) {}
    return null;
  }

}
