/*
 * @(#)HStringTokenizer.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.util;

import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.List;
import java.util.StringTokenizer;

/**
 * A tokenizer with a hierarchy of tokens, used to parse a String with groups
 * and operations.
 * <br><br>
 * The format of the string to parse is: "expression" or "expression1,
 * expression2, ...".<br>
 * <br>
 * An expression is composed of operations between expressions, groups and/or
 * functions. It can use the '=' sign as a group separator.<br>
 * An operation is defined by the use of one of these operators: '+', '-', '*'
 * and '/'.<br>
 * A group is defined by the use of surrounding '(' and ')', '[' and ']', or '{'
 * and '}'.<br>
 * A function is the same as a group, but is prefixed by a function name.<br>
 * <br>
 * The tokenization priorities , from highest to lowest, are: ',', '=', '*' and
 * '/', '+' and '-'. Groups and functions are processed as single expressions,
 * which are themselves tokenized.
 * <br><br>
 * The tokens that are generated reflect the groups and operations. The root
 * token contains the full text as a token, and no operator (with value 0 that
 * is). Its subtokens (from 0 to n) reflect the expressions.<br>
 * The <code>getOperator()</code> indicates the operation, group or function
 * that links subtokens. The operator is one of '=', '+', '-', '*', '/', '(',
 * '[' or '{'. For functions, the <code>getToken()</code> returns the name of
 * the function.<br>
 * <br>
 * For example "x=5+3*max(2, min(1, 3)), -5*-3" is decomposed into the
 * following tree of tokens (notation is "operator, text"):<br>
 * <pre>
 * 0, "x=5+3*max(2, min(1, 3)), -5*-3"
 *   '=', null
 *     0, "x"
 *     '+', null
 *       0, "5"
 *       '*', null
 *         0, "3"
 *         '(', "max"
 *           0, "2"
 *           '(', "min"
 *             0, "1"
 *             0, "3"
 *   '*', null
 *     '-', null
 *       (null token)
 *       0, "5"
 *     '-', null
 *       (null token)
 *       0, "-3"
 * </pre>
 * @version 1.0 2004.01.06
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class HStringTokenizer {

  /** The list of groups. */
  protected List groupList;

  /** The text contained in this expression. */
  protected String text;

  /** The operator. */
  protected char operator = 0;

  /**
   * Construct a tokenizer for the particular text.
   * @param text The text to break into grouped tokens.
   */
  public HStringTokenizer(String text) {
    this(text, SPLIT_GROUPS);
    if(operator != ',') {
      if(operator != 0 || text != null && text.trim().length() != 0) {
        HStringTokenizer sub = new HStringTokenizer(this.text, -1);
        sub.operator = operator;
        sub.groupList = groupList;
        groupList = new ArrayList(1);
        groupList.add(sub);
      }
    }
    this.text = text;
    operator = 0;
  }

  /**
   * Get the operator.
   * The operator is one of:
   * <ul>
   *   <li>',' or '=' for basic group separation. A ',' only appears for the
   * root of the text that was split.</li>
   *   <li>'[', '(' or '{' for group encapsulation (implicitely meaning
   * corresponding ']', ')' or '}'). The token for a group may not be null and
   * in that case is the name of a function.
   *   <li>'+', '-', '*', '/' for operations.</li>
   *   <li>0 otherwise.</li>
   * </ul>
   * @return The operator, or 0 if not defined.
   */
  public char getOperator() {
    return operator;
  }

  /**
   * Get the token.
   * @return The token, or null if sub tokens contain the information.
   */
  public String getToken() {
    return text;
  }

  /**
   * Get the sub tokenizers. For a group that contain different comma-separated
   * elements, the elements are direct sub tokenizers.
   * @return The sub tokenizers, or null if this tokenizer contains the
   * information.
   */
  public HStringTokenizer[] getSubTokenizers() {
    if(groupList == null) {
      return new HStringTokenizer[0];
    }
    HStringTokenizer[] tokens = new HStringTokenizer[groupList.size()];
    for(int i=0; i<tokens.length; i++) {
      tokens[i] = (HStringTokenizer)groupList.get(i);
    }
    return tokens;
  }

  /**
   * Construct a group token for some existing group list.
   * @param groupList The group list.
   */
  protected HStringTokenizer(List groupList) {
    this.groupList = groupList;
    process(SPLIT_EQUAL+1);
  }

  /**
   * Construct a tokenizer for the particular text, already in a group.
   * @param groupText The text of the group that contains the text.
   * @param text The text to break into grouped tokens.
   */
  protected HStringTokenizer(String groupText, String text) {
    this.text = groupText;
    if(text.length() != 0) {
      groupList = new ArrayList(1);
      groupList.add(new HStringTokenizer(text, SPLIT_GROUPS));
    }
    process(SPLIT_GROUPS + 1);
    if(groupList != null && groupList.size() == 1) {
      HStringTokenizer hst = (HStringTokenizer)groupList.get(0);
      if(hst.operator == ',') {
        groupList = hst.groupList;
      }
    }
  }

  /**
   * Construct a tokenizer for the particular text.
   * @param text The text to break into grouped tokens.
   * @param step The step from which to start.
   */
  protected HStringTokenizer(String text, int step) {
    if(text != null) {
      this.text = text.trim();
    }
    process(step);
  }

  /**
   * Split the groups according to "name "["[]", "{}", "()", "<>"].
   */
  protected void splitGroups() {
    String text = this.text;
    if(text != null) {
      int copyIndex = 0;
      char foundGroup = 0;
      char endChar = 0;
      int wordStart = -1;
      int count = 0;
      int start = 0;
      for(int i=0; i<text.length(); i++) {
        char currentChar = text.charAt(i);
        switch(currentChar) {
          case '<':
            if(endChar == 0) endChar = '>';
          case '{':
            if(endChar == 0) endChar = '}';
          case '(':
            if(endChar == 0) endChar = ')';
          case '[':
            if(endChar == 0) endChar = ']';
            if(count == 0) {
              foundGroup = currentChar;
              start = i;
            }
            if(foundGroup == currentChar) {
              count++;
            }
            break;
          case '>':
          case '}':
          case ')':
          case ']':
            if(currentChar == endChar) {
              if(count == 1) {
                if(groupList == null) {
                  groupList = new ArrayList();
                  this.text = null;
                }
                String beginText = text.substring(copyIndex, wordStart != -1? wordStart: start).trim();
                if(beginText.length() != 0) {
                  groupList.add(new HStringTokenizer(beginText, -1));
                }
                String group = null;
                if(wordStart != -1) {
                  String word = text.substring(wordStart, start).trim();
                  if(word.length() > 0) {
                    group = word;
                  }
                }
                HStringTokenizer token = new HStringTokenizer(group, text.substring(start+1, i));
                HStringTokenizer newToken = new HStringTokenizer(null, -1);
                newToken.groupList = new ArrayList(1);
                newToken.groupList.add(token);
                if(token.operator == 0) {
                  token.operator = foundGroup;
                } else {
                  newToken.operator = foundGroup;
                }
                newToken.process(SPLIT_GROUPS);
                token = newToken;
                groupList.add(token);
                copyIndex = i + 1;
                endChar = 0;
                wordStart = -1;
              }
              count--;
            }
            break;
          default:
            if(count == 0) {
              if(currentChar == '.' || Character.isLetterOrDigit(currentChar)) {
                if(wordStart == -1) {
                  wordStart = i;
                }
              } else if(!Character.isWhitespace(currentChar)) {
                wordStart = -1;
              }
            }
            break;
        }
      }
      if(count > 0) {
        throw new IllegalStateException("Character \"" + endChar + "\" was not found to match opening \""+ foundGroup + "\"!");
      }
      if(groupList != null) {
        if(copyIndex < text.length()) {
          groupList.add(new HStringTokenizer(text.substring(copyIndex), -1));
        }
      }
    }
  }

  /**
   * Split the separators, which are blanks, operators, etc.
   */
  protected void splitSeparators() {
    if(groupList != null) {
      for(int i=0; i<groupList.size(); i++) {
        HStringTokenizer group = (HStringTokenizer)groupList.get(i);
        String gText = group.text;
        if(gText != null && group.operator == 0) {
          StringTokenizer st = new StringTokenizer(gText, "*/+-,= \t", true);
          int count = 0;
          groupList.remove(i);
          while(st.hasMoreTokens()) {
            String token = st.nextToken().trim();
            if(token.length() > 0) {
              groupList.add(i + count++, new HStringTokenizer(token, SPLIT_SEPARATORS + 1));
            }
          }
          i+=count-1;
        }
      }
    } else {
      StringTokenizer st = new StringTokenizer(text, "*/+-,= \t", true);
      if(st.countTokens() > 1) {
        groupList = new ArrayList();
        while(st.hasMoreTokens()) {
          String token = st.nextToken().trim();
          if(token.length() > 0) {
            groupList.add(new HStringTokenizer(token, SPLIT_SEPARATORS + 1));
          }
        }
        text = null;
      }
    }
  }

  /**
   * Split groups according to commas.
   */
  protected void splitCommas() {
    // Main groups are created at this point. Let's process commas.
    if(groupList != null) {
      List commaPositionsList = new ArrayList();
      for(int i=0; i<groupList.size(); i++) {
        HStringTokenizer group = (HStringTokenizer)groupList.get(i);
        if(group.operator == 0) {
          String gText = group.text;
          if(gText != null && gText.equals(",")) {
            if(i != 0 && i != groupList.size()) {
              commaPositionsList.add(new Integer(i));
            }
            groupList.remove(i--);
          }
        }
      }
      // Need to group all between commas.
      if(!commaPositionsList.isEmpty()) {
        if(((Integer)commaPositionsList.get(commaPositionsList.size() - 1)).intValue() != groupList.size()) {
          commaPositionsList.add(new Integer(groupList.size()));
        }
        if(commaPositionsList.size() > 1) {
          List newGroupList = new ArrayList(commaPositionsList.size());
          int begin = 0;
          for(int i=0; i<commaPositionsList.size(); i++) {
            int end = ((Integer)commaPositionsList.get(i)).intValue();
            if(end - begin == 1) {
              newGroupList.add(groupList.get(begin));
            } else {
              HStringTokenizer group = new HStringTokenizer(null, SPLIT_COMMAS);
              group.groupList = new ArrayList();
              for(int j=begin; j<end; j++) {
                group.groupList.add(groupList.get(j));
              }
//              group.groupList.addAll(groupList.subList(begin, end));
              newGroupList.add(group);
              for(int j=0; j<newGroupList.size(); j++) {
                ((HStringTokenizer)newGroupList.get(j)).process(SPLIT_COMMAS+1);
              }
            }
            begin = end;
          }
          groupList = newGroupList;
          if(operator == 0) {
            operator = ',';
          }
        }
      }
    }
  }

  /**
   * Split groups according to equal.
   */
  protected void splitEqual() {
    // Main groups are created at this point. Let's process commas.
    if(groupList != null) {
      for(int i=0; i<groupList.size(); i++) {
        HStringTokenizer group = (HStringTokenizer)groupList.get(i);
        if(group != null && group.operator == 0) {
          String gText = group.text;
          if(gText != null && gText.equals("=")) {
            groupList.remove(i);
            if(i != 0 && i != groupList.size()) {
              operator = '=';
              int leftSize = i;
              HStringTokenizer leftToken;
              if(leftSize == 1) {
                leftToken = (HStringTokenizer)groupList.get(0);
              } else {
                List leftList = new ArrayList(leftSize);
                for(int j=0; j<i; j++) {
                  leftList.add(groupList.get(j));
                }
//                leftList.addAll(groupList.subList(0, i));
                leftToken = new HStringTokenizer(leftList);
              }
              int rightSize = groupList.size() - leftSize;
              HStringTokenizer rightToken;
              if(rightSize == 1) {
                rightToken = (HStringTokenizer)groupList.get(leftSize);
                rightToken.process(SPLIT_EQUAL + 1);
              } else {
                List rightList = new ArrayList(rightSize);
                for(int j=leftSize; j<groupList.size(); j++) {
                  rightList.add(groupList.get(j));
                }
//                rightList.addAll(groupList.subList(leftSize, groupList.size()));
                rightToken = new HStringTokenizer(rightList);
              }
              groupList = new ArrayList(2);
              groupList.add(leftToken);
              groupList.add(rightToken);
              break;
            }
          }
        }
      }
    }
  }

  /**
   * Extract negative expressions.
   */
  protected void extractNegative() {
    if(groupList != null) {
      boolean isOperator = true;
      for(int i=0; i<groupList.size()-1; i++) {
        HStringTokenizer group = (HStringTokenizer)groupList.get(i);
        if(group != null) {
          String gText = group.text;
          if(isOperator && gText != null && gText.equals("-")) {
            group.operator = '-';
            group.text = null;
            group.groupList = new ArrayList(2);
            group.groupList.add(null);
            HStringTokenizer token = (HStringTokenizer)groupList.remove(i+1);
            group.groupList.add(token);
            token.process(EXTRACT_NEGATIVE + 1);
            group.process(EXTRACT_NEGATIVE + 1);
            isOperator = false;
          } else {
            isOperator = gText != null && (gText.equals("+") || gText.equals("*") || gText.equals("/") || gText.equals("-"));
          }
        }
      }
    }
  }

  /**
   * Split groups according to * and /.
   */
  protected void splitMultiplications() {
    splitOperations(SPLIT_MULTIPLICATIONS, "*", "/");
  }

  /**
   * Split groups according to + and -.
   */
  protected void splitAdditions() {
    splitOperations(SPLIT_ADDITIONS, "+", "-");
  }

  /**
   * Split an operation of a step, using the 2 operators.
   * @param step The current step.
   * @param char1 One of the 2 characters to use as an operator.
   * @param char2 One of the 2 characters to use as an operator.
   */
  protected void splitOperations(int step, String char1, String char2) {
    if(groupList != null) {
      for(int i=0; i<groupList.size()-2; i++) {
        HStringTokenizer group = (HStringTokenizer)groupList.get(i+1);
        String gText = group.text;
        if(gText != null && (gText.equals(char1) || gText.equals(char2))) {
          group.operator = gText.charAt(0);
          group.text = null;
          group.groupList = new ArrayList(2);
          HStringTokenizer left = (HStringTokenizer)groupList.remove(i);
          left.process(step + 1);
          group.groupList.add(left);
          HStringTokenizer right = (HStringTokenizer)groupList.remove(i+1);
          right.process(step + 1);
          group.groupList.add(right);
          group.process(step + 1);
          i--;
        }
      }
    }
  }

  /** Step to split the groups according to "name "["[]", "{}", "()", "<>"]. */
  protected final static int SPLIT_GROUPS = 0;
  /** Step to split groups according to the separators. */
  protected final static int SPLIT_SEPARATORS = 1;
  /** Step to split groups according to commas. */
  protected final static int SPLIT_COMMAS = 2;
  /** Step to split groups according to '='. */
  protected final static int SPLIT_EQUAL = 3;
  /** Step to extract negative expressions. */
  protected final static int EXTRACT_NEGATIVE = 4;
  /** Step to split groups according to '*' and '/'. */
  protected final static int SPLIT_MULTIPLICATIONS = 5;
  /** Step to split groups according to '+' and '-'. */
  protected final static int SPLIT_ADDITIONS = 6;
  /** Step to remove useless intermediate levels. */
  protected final static int PRUNE = 7;

  /**
   * Run the process.
   * @param step The step at which to start the process.
   */
  protected void process(int step) {
    switch(step) {
      case SPLIT_GROUPS:
        splitGroups();
      case SPLIT_SEPARATORS:
        splitSeparators();
      case SPLIT_COMMAS:
        splitCommas();
      case SPLIT_EQUAL:
        splitEqual();
      case EXTRACT_NEGATIVE:
        extractNegative();
      case SPLIT_MULTIPLICATIONS:
        splitMultiplications();
      case SPLIT_ADDITIONS:
        splitAdditions();
      case PRUNE:
        prune();
    }
  }

  /**
   * Remove useless intermediate levels.
   */
  protected void prune() {
    if(groupList != null && groupList.size() == 1) {
      HStringTokenizer group = (HStringTokenizer)groupList.get(0);
      if(group.text == null && group.operator == 0) {
        groupList = group.groupList;
      } else if(text == null && operator == 0) {
        groupList = group.groupList;
        text = group.text;
        operator = group.operator;
      }
    }
  }

}
