/*
 * @(#)DtpPoint.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.unit;

/**
 * A handler to manipulate a default unit representing a DTP point.
 * @version 1.0 2005.02.01
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class DtpPoint {

  /**
   * A unit representing a DTP point.
   * @version 1.0 2005.02.01
   * @author Christopher Deckers (chrriis@brainlex.com)
   */
  public static class DtpPointUnit extends ScreenUnit {
    
    /** The ratio. */
    protected double dtpPointUnitRatio;

    /**
     * Construct a new DtpPointUnit.
     */
    protected DtpPointUnit() {
      dtpPointUnitRatio = super.getRatio() / 72;
    }
    
    /**
     * Get the ratio to use when making conversions. The ratio is so that:<br>
     * <code>unitValue * ratio == pixelValue</code>.
     * @return The ratio to use for conversions.
     */
    protected double getRatio() {
      return dtpPointUnitRatio;
    }

    /**
     * Get the names representing the unit.
     * @return An array containing the names.
     */
    public String[] getNames() {
      return new String[] {"pt", "point"};
    }
  }
  
  /** Hiden constructor of a DtpPoint. */
  protected DtpPoint() {}
  
  /** The default unit that is used. */
  protected static DtpPointUnit unit;
  
  /**
   * Get the default unit.
   * @return The default unit.
   */
  public static DtpPointUnit getUnit() {
    synchronized(DtpPoint.class) {
      if(unit == null) {
        unit = newInstance();
      }
    }
    return unit;
  }
  
  /**
   * Create a new instance of this unit.
   * @return The newly created instance.
   */
  public static DtpPointUnit newInstance() {
    return new DtpPointUnit();
  }

  /**
   * Get the double value resulting from the conversion of the value from the specified unit.
   * @param unitName A name identifying the unit from which to convert.
   * @param value The value to convert.
   * @return The result of the conversion.
   */
  public static double doubleValue(String unitName, double value) {
    Unit unit = Unit.getUnit(unitName);
    if(unit == null) {
      throw new IllegalArgumentException("The unit \""+ unit + "\" is not registered!");
    }
    return getUnit().doubleValue(unit, value);
  }
  
  /**
   * Get the double value resulting from the conversion of the value from the specified unit.
   * @param unit The unit from which to convert.
   * @param value The value to convert.
   * @return The result of the conversion.
   */
  public static double doubleValue(Unit unit, double value) {
    return getUnit().doubleValue(unit, value);
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to this unit.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The int value resulting from the conversion.
   */
  public static int intValue(String unitValue) {
    return (int)Math.round(doubleValue(unitValue));
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to this unit.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The double value resulting from the conversion.
   */
  public static double doubleValue(String unitValue) {
    return doubleValue(Pixel.getUnit(), Unit.parse(unitValue));
  }

}
